// Telegraph Server
// Nate Hunzaker
//
// A framework based on Backbone for event driven applications


//-- Requirements ------------------------------------------------------//

// Side Dishes
require('colors');

var _          = require('underscore'),
express   = require('express'),
ejs             = require('ejs'),
fs               = require('fs');

// The Main Course
module.exports.Telegraph = Telegraph = require('backbone');

// Cleanup modules we'll never use
delete Telegraph.Router;
delete Telegraph.View;

// We want to save the extend method for later
var extend = Telegraph.Model.extend;


//-- Telepathy Server ---------------------------------------------------//
Telegraph.Server = function (options) {
    
    //-- Default Values ---------------------------------------------------//
    var settings = {
        'appname'        : "Telegraph",
        'database'         : false,
        'express'            : express.createServer(),
        'public'               : 'public',
        'routes'              : {},
        'socketio'           : false,
        'views'                : 'views',
        'view engine'    : 'ejs' 
    };

    // Add Default Overides
    _.extend(this, settings, options);

    // Set everything else up
    this._setup();
};


// Backbone Server Functionality
_.extend(Telegraph.Server.prototype, Telegraph.Events, {
             
             //-- Modules  ------------------------------------------------------//

             // A nifty helper to load all files within a directory
             // Note : This is basically a glorified "require all in directory" function.
             //        Should further releases of node include such functionality, it will
             //        be deprecated.
             load: function(module) {
                 
                 var dir = require.main.filename.split("/").slice(0, require.main.filename.split("/").length - 1).join("/"),
                 files = fs.readdirSync( dir + "/" + module, 'utf-8');
                 
                 // Filter hidden files and folders
                 files = files.filter(function(file) {
                                          return (/^[^.].+\..+/g.test(file) === true);
                                      });

                 files.forEach(function(file) {
                                   require( dir + "/" + module + "/" + file );
                               });

             },
             
             //-- Standard CRUD Routing Methods  --------------------------------//

             route: function (method, path, action) {
                 this.routes[path] = [method, action];
             },

             get: function(path, action) {
                 this.route('get', path, action);
             },

             post: function(path, action) {
                 this.route('post', path, action);
             },

             put: function(path, action) {
                 this.route('put', path, action);
             },

             delete: function(path, action) {
                 this.route('delete', path, action);
             },

             _setup: function() {
                 
                 var server = this;

                 // Set static content folder
                 this.express.use(express.static(this.public));
                 
                 // Set the view engine
                 this.express.set('view engine', this['view engine']);    

                 //-- SocketIO Methods  -----------------------------------------//
                 
                 if ( this.socketio ) {

                     this.io = require('socket.io').listen(this.express);
                     
                     // A syntax sugar for socketIO
                     this.listen = function(name, action) {
                         this.io.sockets.on(name, action);
                     };
                     
                     this.send = function(name, values) {
                         this.io.sockets.emit(name, values);
                     };

                     this.events = function(events) {
                         
                         for (var e in events) {
                             this.bind(e, events[e]);
                         }

                     };

                 }

                 //-- Database Methods  --------------------------------------------//
                 
                 if ("database" in this) {            
                     require("./telegraph-collection");
                     require('./telegraph-database').setup(this.database);
                 } 

                 // Set up inheritance for the model and collection
                 Telegraph.Model.extend = Telegraph.Collection.extend = extend;
             },


             //-- Server Methods ------------------------------------------------//

             start: function(port) {   

                 // Add routes
                 for (var i in this.routes) {
                     
                     var method = this.routes[i][0],
                     action = this.routes[i][1];
                     
                     this.express[method](i, action);
                 }
                 
                 this.express.listen(port);

                 console.log("\n" + this.appname.toString().cyan.bold + " is listening on port " + port.toString().cyan.bold);
             }
             
         });
