// Telepathy.Collection
// -------------------

var _ = require("underscore"),
mongoose = require('mongoose'),
Schema = mongoose.Schema;

// Provides a standard collection class for our sets of models, ordered
// or unordered. If a `comparator` is specified, the Collection will maintain
// its models in sort order, as they're added and removed.
Telegraph.Collection = function(options) {

    options || (options = {});
    
    _.extend(this, options);
    
    _.bindAll(this, '_onModelEvent', '_removeReference');
    
    // Generate Mongo Relationship
    var new_schema = new Schema(this.schema);
    this._datalink = mongoose.model(this.name, new_schema);

    var m = this.model = new Telegraph.Model();
    m.defaults = {};
    
    for (var d in this.schema) {
        m.defaults[d] = this.schema[d].default;   
    }

    this.initialize.apply(this, arguments);
};

// Define the Collection's inheritable methods.
_.extend(Telegraph.Collection.prototype, Telegraph.Events, {

             // The default model for a collection is just a **Backbone.Model**.
             // This should be overridden in most cases.
             model : Telegraph.Model,

             // Initialize is an empty function by default. Override it with your own
             // initialization logic.
             initialize : function() {},

             // The JSON representation of a Collection is an array of the
             // models' attributes.
             toJSON : function() {
                 return this.map(function(model){ return model.toJSON(); });
             },

             // Add a model, or list of models to the set. Pass **silent** to avoid
             // firing the `added` event for every new model.
             add : function(models, options) {
                 if (_.isArray(models)) {
                     for (var i = 0, l = models.length; i < l; i++) {
                         this._add(models[i], options);
                     }
                 } else {
                     this._add(models, options);
                 }
                 return this;
             },

             // Remove a model, or a list of models from the set. Pass silent to avoid
             // firing the `removed` event for every model removed.
             remove : function(models, options) {
                 if (_.isArray(models)) {
                     for (var i = 0, l = models.length; i < l; i++) {
                         this._remove(models[i], options);
                     }
                 } else {
                     this._remove(models, options);
                 }
                 return this;
             },

             // Get a model from the set by id.
             get : function(id) {
                 if (id == null) return null;
                 return this._byId[id.id != null ? id.id : id];
             },

             // Get the model at the given index.
             at : function(index) {
                 return this.models[index];
             },

             // Force the collection to re-sort itself. You don't need to call this under normal
             // circumstances, as the set will maintain sort order as each item is added.
             sort : function(options) {
                 options || (options = {});
                 if (!this.comparator) throw new Error('Cannot sort a set without a comparator');
                 this.models = this.sortBy(this.comparator);
                 if (!options.silent) this.trigger('reset', this, options);
                 return this;
             },

             // Pluck an attribute from each model in the collection.
             pluck : function(attr) {
                 return _.map(this.models, function(model){ return model.get(attr); });
             },

             // When you have more items than you want to add or remove individually,
             // you can reset the entire set with a new list of models, without firing
             // any `added` or `removed` events. Fires `reset` when finished.
             reset : function(models, options) {
                 models  || (models = []);
                 options || (options = {});
                 this.each(this._removeReference);
                 this._reset();
                 this.add(models, {silent: true, parse: options.parse});
                 if (!options.silent) this.trigger('reset', this, options);
                 return this;
             },

             // Fetch the default set of models for this collection, resetting the
             // collection when they arrive. If `add: true` is passed, appends the
             // models to the collection instead of resetting.
             fetch : function(options) {
                 options || (options = {});
                 if (options.parse === undefined) options.parse = true;
                 var collection = this;
                 var success = options.success;
                 options.success = function(resp, status, xhr) {
                     collection[options.add ? 'add' : 'reset'](collection.parse(resp, xhr), options);
                     if (success) success(collection, resp);
                 };
                 options.error = (options.error, collection, options);
                 return (this.sync || Telegraph.sync).call(this, 'read', this, options);
             },

             // Get a model from the set by client id.
             getByCid : function(cid) {
                 return cid && this._byCid[cid.cid || cid];
             },

             // Create a new instance of a model in this collection. After the model
             // has been created on the server, it will be added to the collection.
             // Returns the model, or 'false' if validation on a new model fails.
             create : function(model, options) {
                 var coll = this;
                 options || (options = {});
                 model = this._prepareModel(model, options);
                 if (!model) return false;
                 var success = options.success;
                 options.success = function(nextModel, resp, xhr) {
                     coll.add(nextModel, options);
                     if (success) success(nextModel, resp, xhr);
                 };
                 model.save(null, options);
                 return model;
             },

             // **parse** converts a response into a list of models to be added to the
             // collection. The default implementation is just to pass it through.
             parse : function(resp, xhr) {
                 return resp;
             },

             // Proxy to _'s chain. Can't be proxied the same way the rest of the
             // underscore methods are proxied because it relies on the underscore
             // constructor.
             chain : function () {
                 return _(this.models).chain();
             },

             // Reset all internal state. Called when the collection is reset.
             _reset : function(options) {
                 this.length = 0;
                 this.models = [];
                 this._byId  = {};
                 this._byCid = {};
             },

             // Prepare a model to be added to this collection
             _prepareModel : function(model, options) {
                 
                 // Is the model argument already a Telegraph.Model?
                 if (!(model instanceof Telegraph.Model)) {
                     // no : make it in to one
                     var attrs = model;
                     model = new Telegraph.Model(attrs, { defaults: this.model.defaults, collection: this, parse: options.parse });
                     if (model.validate && !model._performValidation(model.attributes, options)) model = false;

                 } else if (!model.collection) {
                     
                     // yes : just change the reference the collection it belongs to
                     model.collection = this;
                 }

                 return model;
             },

             // Internal implementation of adding a single model to the set, updating
             // hash indexes for `id` and `cid` lookups.
             // Returns the model, or 'false' if validation on a new model fails.
             _add : function(model, options) {
                 options || (options = {});

                 model = this._prepareModel(model, options);

                 if (!model) return false;
                 var already = this.getByCid(model);
                 if (already) throw new Error(["Can't add the same model to a set twice", already.id]);

                 this._byId[model.id] = model;
                 this._byCid[model.cid] = model;

                 var index = options.at != null ? options.at :
                     this.comparator ? this.sortedIndex(model, this.comparator) :
                     this.length;

                 this.models.splice(index, 0, model);

                 model.bind('all', this._onModelEvent);
                 this.length++;

                 options.index = index;

                 if (!options.silent) model.trigger('add', model, this, options);
                 return model;
             },

             // Internal implementation of removing a single model from the set, updating
             // hash indexes for `id` and `cid` lookups.
             _remove : function(model, options) {
                 options || (options = {});
                 model = this.getByCid(model) || this.get(model);
                 if (!model) return null;
                 delete this._byId[model.id];
                 delete this._byCid[model.cid];
                 var index = this.indexOf(model);
                 this.models.splice(index, 1);
                 this.length--;
                 options.index = index;
                 if (!options.silent) model.trigger('remove', model, this, options);
                 this._removeReference(model);
                 return model;
             },

             // Internal method to remove a model's ties to a collection.
             _removeReference : function(model) {
                 if (this == model.collection) {
                     delete model.collection;
                 }
                 model.unbind('all', this._onModelEvent);
             },

             // Internal method called every time a model in the set fires an event.
             // Sets need to update their indexes when models change ids. All other
             // events simply proxy through. "add" and "remove" events that originate
             // in other collections are ignored.
             _onModelEvent : function(ev, model, collection, options) {
                 if ((ev == 'add' || ev == 'remove') && collection != this) return;
                 if (ev == 'destroy') {
                     this._remove(model, options);
                 }
                 if (model && ev === 'change:' + model.idAttribute) {
                     delete this._byId[model.previous(model.idAttribute)];
                     this._byId[model.id] = model;
                 }
                 this.trigger.apply(this, arguments);
             }
             
         });

// Underscore methods that we want to implement on the Collection.
var methods = ['forEach', 'each', 'map', 'reduce', 'reduceRight', 'find', 'detect',
               'filter', 'select', 'reject', 'every', 'all', 'some', 'any', 'include',
               'contains', 'invoke', 'max', 'min', 'sortBy', 'sortedIndex', 'toArray', 'size',
               'first', 'rest', 'last', 'without', 'indexOf', 'lastIndexOf', 'isEmpty', 'groupBy'];

// Mix in each Underscore method as a proxy to `Collection#models`.
_.each(methods, function(method) {
           Telegraph.Collection.prototype[method] = function() {
               return _[method].apply(_, [this.models].concat(_.toArray(arguments)));
           };
       });